#include "wav_file.h"
#include <limits.h>

#define MAX_ 32767

static void write_ushort_(File *file, ushort u);

static void write_short_(File *file, short s);

static void write_uint_(File *file, uint u);

static ushort read_ushort_(File *file);

static short read_short_(File *file);

static uint read_uint_(File *file);

static void confirm_ushort_(File *file, ushort expected);

static void confirm_uint_(File *file, uint expected);


DoubleAudio *double_audio_new_load_wav(DoubleAudio **p_this, char *filename) {
  uint bytes_1;
  uint bytes_2;
  uint n;
  uint rate;
  uint u;
  double *samples;
  File *_f = null;
  DoubleAudio *this;

  file_new(&_f, filename, "r");

  confirm_uint_(_f, 0x46464952);

  bytes_1 = (read_uint_(_f) - 36);

  confirm_uint_(_f, 0x45564157);
  confirm_uint_(_f, 0x20746d66);
  confirm_uint_(_f, 0x10);
  confirm_ushort_(_f, 0x01);
  confirm_ushort_(_f, 0x01);
  rate = read_uint_(_f);
  confirm_uint_(_f, 2 * rate);
  confirm_ushort_(_f, 0x02);
  confirm_ushort_(_f, 0x10);
  confirm_uint_(_f, 0x61746164);

  bytes_2 = read_uint_(_f);

  if (bytes_2 != bytes_1) {
    throw("Expected %u but got %u", bytes_1, bytes_2);
  }

  n = bytes_1 / 2;

  double_audio_new(p_this, n, rate);
  this = *p_this;

  samples = this->_samples;

  for (u = 0; u < n; u++) {
    samples[u] = (double)read_short_(_f) / MAX_;
  }

  file_delete(&_f);

  return this;
}

void double_audio_save_wav(DoubleAudio *this, char *filename) {
  short s;
  uint u;
  ulong n;
  double error;
  double v;
  double *samples;
  File *_f = null;

  n = this->n;
  samples = this->_samples;

  if (n > UINT_MAX) {
    throw("Number of samples (%lu) too large for WAV file format (max %lu)",
        n,
        UINT_MAX);
  }

  file_new(&_f, filename, "w");

  file_write_string(_f, "RIFF");
  write_uint_(_f, 36 + 2 * n);
  file_write_string(_f, "WAVE");
  file_write_string(_f, "fmt ");
  write_uint_(_f, 16);
  write_ushort_(_f, 1);
  write_ushort_(_f, 1);
  write_uint_(_f, this->rate);
  write_uint_(_f, 2 * this->rate);
  write_ushort_(_f, 2);
  write_ushort_(_f, 16);
  file_write_string(_f, "data");
  write_uint_(_f, 2 * n);

  error = 0;

  for (u = 0; u < n; u++) {
    v = error + samples[u];
    v = bound(v, -1, 1);
    s = (short)round(v * MAX_);
    error = v - (double)s / MAX_;

    if (error < -0.5 / MAX_ || error > +0.5 / MAX_) {
      throw("impossible");
    }

    write_short_(_f, s);
  }

  file_delete(&_f);
}


static void confirm_ushort_(File *file, ushort expected) {
  ushort u;

  if ((u = read_ushort_(file)) != expected) {
    throw("Expected ushort %d but read %d", expected, u);
  }
}

static void write_ushort_(File *file, ushort u) {
  file_write_byte(file, (byte)(u & 0xff));
  file_write_byte(file, (byte)((u >> 8) & 0xff));
}

static void write_uint_(File *file, uint u) {
  write_ushort_(file, u & 0xffff);
  write_ushort_(file, (u >> 16) & 0xffff);
}

static ushort read_ushort_(File *file) {
  return (ushort)file_read_byte(file) | (ushort)file_read_byte(file) << 8;
}

static uint read_uint_(File *file) {
  return (uint)read_ushort_(file) | (uint)read_ushort_(file) << 16;
}

static void confirm_uint_(File *file, uint expected) {
  uint u;

  if ((u = read_uint_(file)) != expected) {
    throw("Expected uint 0x%x but read 0x%x", expected, u);
  }
}

static void write_short_(File *file, short s) {
  write_ushort_(file, (ushort)s);
}

static short read_short_(File *file) {
  return (short)read_ushort_(file);
}
